/*
 * ConnUnity.cpp - library for the direct Connection to Unity
 *
 * based on code for a direct connection to Unity, and as such provided in the zip file
 * "Connecting Arduino to Unity.zip" by Tiuri de Jong
 * https://code.google.com/p/unity-arduino-serial-connection/
 *
 *
 * Adapted by Gidi van Liempd (gidi@geedesign.com) Jan 2013
*/

//******************************************************************************
//* Includes
//******************************************************************************

#include "ConnUnity.h"

extern "C" {
#include <string.h>
#include <stdlib.h>
}

//******************************************************************************
//* Constructors
//
// 	boolean bFakeDigital = TRUE if we use only fake digital inputs (i.e., pass the
// 		values of other computations as if it were the values of an actual pin)
//	boolean bFakeAnalog = TRUE if we use only fake analog inputs
//
// Make sure that the pinConfig Array passed as an argument is exactly the same
// as in the Unity files.
//******************************************************************************

ConnUnity::ConnUnity(boolean bFakeDigital, boolean bFakeAnalog, char pinConfig[])
{
	stringLength = 1;
	digOut = 0;
	digIn = 0;
	//
	_bFakeDigital = bFakeDigital;
	_bFakeAnalog = bFakeAnalog;

 	// set pinconfiguration according to the pinConfig array passed by reference
  	for (int i =0; i < MAX_DIGITAL; i++) {	// Gidi: notice we have a problem if the array length is off
		_pinConfig[i] = pinConfig[i];
	}
	// (Gidi) If we use virtual pins, the next step would be unnecessary
	// However, in the communication with Flash/Unity this configuration is still checked.
	// So we leave it as it is, (and make sure we set actual pins AFTER this Constructor
	//
	setupPins();	// Function to setup the (physical) digital pin configuration
}

//******************************************************************************
//* Public Methods
//******************************************************************************

// -----------------------------------------------------------------------
// (Gidi) Function to read a (virtual) digital pin value
//
// pin = virtual pin number
// N.B. Notice that this only works correctly if we assign the input pins
// consecutively!
// -----------------------------------------------------------------------
void ConnUnity::readVirtualDigital(byte pin, boolean value)
{
	digInLastValues[pin] = value;
}

// -----------------------------------------------------------------------
// (Gidi) Function to read a (virtual) analog pin value
//
// pin = virtual pin number
// -----------------------------------------------------------------------
void ConnUnity::readVirtualAnalog(byte pin, int value)
{
	analogInLastValues[pin] = value;
}

// -----------------------------------------------------------------------
// Function to process serial input and output
//
// -----------------------------------------------------------------------
void ConnUnity::processSerial()
{

  // read the serial port and create a string out of what you read
  readSerialString(serInString);

  // check if there data is received if isStringEmpty is true do nothing.
  if(isStringEmpty(serInString) == false) {

    /*
     check if the received string starts with an o
     check if string length is the same as expected length (defined outputs)
     if true : return I (input), send input info to Flash and update the Arduino outputs
     if false: return E (error) and the part of the string that is received (to debug)

     check if the received string start with an n
     send only input information ( no new info from Flash )
     */

    if(serInString[0]==111) //o
    {
      if(serInStringLength>=stringLength)
      { //the received information is complete
        updateArduinoOutputs();
        sendInputsToFlash(false);
      }
      // receive info incomplete; error is true
      else sendInputsToFlash(true);
    }
    else if(serInString[0]==110) //n
    {
      sendInputsToFlash(false);
    }
    // no o or n as start character, start with error info
    else sendInputsToFlash(true);

    eraseStrings();
  }

}

//******************************************************************************
//* Private Methods
//******************************************************************************

// -----------------------------------------------------------------------
// Function to setup the DIGITAL pin configuration
// -----------------------------------------------------------------------
void ConnUnity::setupPins()
{
  // set pinconfiguration according to the _pinConfig array
  for (int i =0; i < MAX_DIGITAL; i++) {
    //Serial.print(_pinConfig[i]);
    if (_pinConfig[i] == 'i') { // input
		if (!_bFakeDigital) {
      		pinMode(i + PIN_OFFSET, INPUT);
		}
      	digInputs[digIn] = i + PIN_OFFSET;
      	digIn++;
    }
    if (_pinConfig[i] == 'o') { // output
		if (!_bFakeDigital) {
      		pinMode(i + PIN_OFFSET, OUTPUT);
		}
      	digOutputs[digOut] = i + PIN_OFFSET;
      	stringLength +=2; // output string part will have 2 characters
      	digOut++;
    }
    if (_pinConfig[i] == 'p') { // pwm output.
		if (!_bFakeDigital) {
      		pinMode(i + PIN_OFFSET, OUTPUT);
		}
      	digOutputs[digOut] = i + PIN_OFFSET;
      	stringLength +=4; // pwm string part will have 4 characters
      	digOut++;
    }
    // Notice: pins which do not have a i,o or p value are ignored!
    // (Can be used to ignore RX/TX on Teensy2.0??)
  }
}

// -----------------------------------------------------------------------
// Function to read a string from the serialport and store it in an array
// -----------------------------------------------------------------------
void ConnUnity::readSerialString (char *strArray)
{
  int i = 0;
  if(Serial.available()){
    while (Serial.available()) {
      strArray[i] = Serial.read();
      i++;
    }
  }
  serInStringLength=i;
}

// -----------------------------------------------------------------------
// Function to find out wether the array is empty or not
// -----------------------------------------------------------------------
boolean ConnUnity::isStringEmpty(char *strArray)
{
  if (strArray[0] == 0) return true;
  else 			return false;
}

// -----------------------------------------------------------------------
// Function which sends Arduino input values to Flash/Unity
//
// error = whether or not an error was detected in the serial string coming
//	from Flash/Unity
// -----------------------------------------------------------------------
void ConnUnity::sendInputsToFlash(boolean error)
{
  if(error) {
    Serial.print("e,");
    Serial.print(stringLength);
    Serial.print(",");
    Serial.print(serInStringLength);
    Serial.print(",");
  }
  else  Serial.print("i,");

  //
  // read the defined digital inputs
  //
  for (int i=0; i < digIn; i++) {
    if (_bFakeDigital) {	// _bFakeDigital = whether or not to use the actual pins, or fake inputs
    	val = digInLastValues[i];
    }
    else {	// read the actual pin
    	val = digitalRead(digInputs[i]);
    };
    // process val
    if (val == HIGH) {
      Serial.print(val,BIN);
    }
    if (val == LOW) {
      Serial.print(val,BIN);
    }
    Serial.print(",");
  }

  //
  // read the analog inputs
  //
  for (int i=0; i < MAX_ANALOG; i++) {
    if (_bFakeAnalog) {	// _bFakeAnalog = whether or not to use the actual pins, or fake inputs
    	val = analogInLastValues[i];
    }
    else {	// read the actual pin
    	val = analogRead(i);
    };
    ///*
    if(i != (MAX_ANALOG - 1)){
      Serial.print(val,DEC);
      Serial.print(",");
    }else{
      Serial.println(val,DEC);
    }
    //*/
    //Serial.print(val,DEC);
    //Serial.print(",");
  }

  //Serial.println(0, BYTE); // close with a zero byte for XML Flash
  // Commented out, because Unity does NOT like these zero bytes
}

// -----------------------------------------------------------------------
// Function to update the Arduino outputs, with values received from Flash/Unity
// -----------------------------------------------------------------------
void ConnUnity::updateArduinoOutputs()
{ // position of the current character in the serInstring
  //int serInStringPosition = 1;
  int serInStringPosition = 2;

  // set the defined outputs

  for (int i=0; i < digOut; i++) {

    // check if next character is ',' then it is a digital output
    // because a pwm value has three characters

    if (serInString[serInStringPosition + 1]==',' ) {
      tempString[0] = serInString[serInStringPosition];
      tempString[1] = 0;
      tempString[2] = 0;
      tempVal = atoi(tempString);

      // check if this output has changed
      // this prevents outputs to "flicker"
      if(digOutLastValues[i]!=tempVal)
      {
        if(tempVal==1) digitalWrite(digOutputs[i], HIGH);
        if(tempVal==0) digitalWrite(digOutputs[i], LOW);
        digOutLastValues[i]=tempVal;
      }
      serInStringPosition += 2;
    }

    // next pin is not ',' so its a pwm value
    else {
      tempString[0] = serInString[serInStringPosition];
      tempString[1] = serInString[serInStringPosition+1];
      tempString[2] = serInString[serInStringPosition+2];
      tempVal = atoi(tempString);

      if(digOutLastValues[i]!=tempVal)
      {
        analogWrite(digOutputs[i],tempVal);
        digOutLastValues[i]=tempVal;
      }
      serInStringPosition +=4;
    }
  }

}

// -----------------------------------------------------------------------
// Function to erase contents in serInString array
// -----------------------------------------------------------------------
void ConnUnity::eraseStrings()
{ // erase contents in the serInString array
  for (int i=0; i < ARRAY_LENGTH; i++) {
    serInString[i]=0;
  }
  // erase contents in the tempString array
  for (int i=0; i < 4; i++) {
    tempString[i]=0;
  }
}

